/*
 * $FreeBSD$
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <libnet.h>

#define LEET 31337

/*
 * Packet structures and defines.
 */

struct gre_h {
        u_int16_t flags;
        u_int16_t ptype;
#ifdef USE_GRE_CKSUM
	u_int16_t gcksum;
	u_int16_t pad00;
#endif /* USE_GRE_CKSUM */
};
struct greip {
        struct ip gi_i;
        struct gre_h  gi_g;
};
struct pkt { 
	struct libnet_ip_hdr	ip;
	struct gre_h		gre;
	struct libnet_ip_hdr	ipip;
	struct libnet_udp_hdr	udp;
	u_char			buf[48];
};

#define gi_pr           gi_i.ip_p
#define gi_len          gi_i.ip_len
#define gi_src          gi_i.ip_src
#define gi_tun_dest     gi_i.ip_tun_dest
#define gi_ptype        gi_g.ptype
#define gi_flags        gi_g.flags
#define gi_cksum        gi_g.gcksum
#define GRE_CP          0x8000  /* Checksum Present */

/* 
 * endpoints to generate packet with.
 * XXX: don't be so lazy and make them user-configurable.
 */
static const char tun_src_str[]		= "172.16.1.2";
static const char tun_dest_str[]	= "172.16.1.1";

static const char message[] = "Aye carumba!";

/*
 * function main().
 */
int
main(int argc, char *argv[])
{
	int s, err, iplen;
	struct pkt	gip;
	struct in_addr	tun_src;
	struct in_addr	phys_dest;
	struct in_addr	tun_dest;
	char		*phys_dest_str;

	if (argc != 2) {
		fprintf(stderr, "usage: greping [dest]\n"
			"dest must be an IP address, not a hostname.\n");
		exit(EXIT_FAILURE);
	}

	/*
	 * no need to look up own address - ip_output()
	 * in the kernel is smart enough to do it for us
	 * and in an interface-independent manner.
	 */

	/*
	 * physical IP target of tunnel
	 */
	phys_dest_str = argv[1];
 	phys_dest.s_addr = inet_addr(phys_dest_str);

	/*
	 * tunnel address endpoints
	 */
 	tun_src.s_addr = inet_addr(tun_src_str);
 	tun_dest.s_addr = inet_addr(tun_dest_str);

	s = libnet_open_raw_sock(IPPROTO_IP);
	if (s == -1) {
		perror("libnet_raw_sock");
		exit(EXIT_FAILURE);
	}
	
	/* build packet from outside in. */
	iplen = sizeof(gip);
	bzero(&gip, iplen);
	libnet_build_ip(iplen - LIBNET_IP_H,
		0, 0, 0, 255, IPPROTO_GRE, INADDR_ANY, phys_dest.s_addr,
		NULL,
		0,
		(void *)&gip.ip);
#ifdef USE_GRE_CKSUM 
	gip.gre.flags = htons(GRE_CP);
#else
	gip.gre.flags = 0;
#endif /* USE_GRE_CKSUM */
	gip.gre.ptype = htons(ETHERTYPE_IP);
	bcopy(message, &gip.buf[0], sizeof(message));
        libnet_build_ip(LIBNET_UDP_H+48,
                IPTOS_LOWDELAY|IPTOS_THROUGHPUT,
                LEET, 0, 255, IPPROTO_UDP, tun_src.s_addr, tun_dest.s_addr,
                (void *)&gip.udp, LIBNET_UDP_H+48,
                (void *)&gip.ipip);
	libnet_build_udp(6969, 6969, (void *) &gip.buf,
		48, (void *) &gip.udp);
	/*
	 * checksum the inside of the tunneled packet.
	 * we do not need to compute outer shell's checksum, as
	 * ip_output() does it for us. udp packets without
	 * checksum (NULL checksum) are legal.
	 */
	libnet_do_checksum((void*)&gip.udp, IPPROTO_UDP, LIBNET_UDP_H+48);
	libnet_do_checksum((void*)&gip.ipip, IPPROTO_IP,
		LIBNET_IP_H);
#ifdef USE_GRE_CKSUM 
	gip.gre.gcksum = 0;
	gip.gre.gcksum = libnet_in_cksum((void *)&gip, sizeof(gip));
	gip.gre.gcksum = LIBNET_CKSUM_CARRY(gip.gre.gcksum);
#endif /* USE_GRE_CKSUM */
	
	err = libnet_write_ip(s, (void*) &gip, iplen);
	usleep(50000);

	if (err == -1) perror("libnet_write_ip");
	libnet_close_raw_sock(s);

	exit(EXIT_SUCCESS);
}
