/*	$FreeBSD$ */

/*
 * Copyright (c) 2001 Organic Systems.
 * All rights reserved
 *
 * This code is conceptually based on the work of the WIDE project
 * and earlier code by Heiko W. Rupp.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by Organic
 *        Systems and its affiliates..
 * 4. Neither the name of Organic Systems nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ORGANIC SYSTEMS AND AFFILIATES
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <sys/syslog.h>
#include <sys/protosw.h>
#include <sys/sysctl.h>
#include <sys/proc.h>
#include <machine/cpu.h>
#include <machine/bus.h>
#include <sys/rman.h>

#include <net/if.h>
#include <net/if_types.h>
#include <net/netisr.h>
#include <net/route.h>
#include <net/bpf.h>
#include <net/ethernet.h>

#define _IP_VHL

#ifdef INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ipprotosw.h>
#endif  /* INET */

#include <net/if_gre.h>
#include <net/net_osdep.h>

/*
 * global variables 
 */

static MALLOC_DEFINE(M_GRE, "gre", "GRE Tunnel Interface");
TAILQ_HEAD(grehead, gre_softc) gres = TAILQ_HEAD_INITIALIZER(gres);

struct if_clone gre_cloner =
	IF_CLONE_INITIALIZER("gre", gre_clone_create, gre_clone_destroy);

static struct rman greunits[1];
static int loaded = 0;

/*
 * static prototypes
 */
static int if_gre_modevent __P((module_t mod, int cmd, void *arg));
static int gre_load __P((void));
static void gre_install_input_hook __P((void));
static void gre_remove_input_hook __P((void));
static int gre_unload __P((void));

/*
 * sysctl-wrapped variables to control acceptance of GRE and Mobile IP
 * traffic destined for this machine. 0 implies deny all, otherwise accept.
 *
 * unlike Solaris, no further stirring is required - kld takes
 * care of these. smart.
 */
int gre_allow = 1;
SYSCTL_INT(_net_inet_ip, OID_AUTO, gre_allow, CTLFLAG_RW,
	   &gre_allow, 0, "Process incoming GRE traffic");
int gre_nest_limit = GRE_NEST_LIMIT;
SYSCTL_INT(_net_inet_ip, OID_AUTO, gre_nest_limit, CTLFLAG_RW,
	   &gre_nest_limit, 0, "Maximum depth of nested GRE tunnels");
int gre_default_mtu = GRE_DEFAULT_MTU;
SYSCTL_INT(_net_inet_ip, OID_AUTO, gre_default_mtu, CTLFLAG_RW,
	   &gre_default_mtu, 0, "Default MTU for GRE tunnels");
#if 0
int ip_mobile_allow = 0;
SYSCTL_INT(_net_inet_ip, OID_AUTO, ip_mobile_allow, CTLFLAG_RW,
	   &ip_mobile_allow, 0, "Process incoming Mobile IP traffic");
#endif


/*
 * protosw structure for hooking into the IP stack.
 * for the curious, /src/sys/netinet/in_proto.c has the protosw
 * tables for the standard IP protocols. use of the term 'rip'
 * here refers to raw IP input/output.
 */
extern struct protosw inetsw[];

#ifdef GRE_EVIL_INPUT_HOOK_HACK
/*
 * an entire protosw[] structure to save RSVP's handlers.
 */
struct protosw rsvp_protosw_save = { 0, };
extern  struct domain inetdomain;
struct ipprotosw gre_protosw =
{ SOCK_RAW,		&inetdomain,	IPPROTO_GRE,	PR_ATOMIC|PR_ADDR,
  gre_input,		rip_output,	NULL,		rip_ctloutput,
  NULL,
  NULL,			NULL,		NULL,		NULL,
  &rip_usrreqs
};
#else
/*
 * simply save the gre_input stub when we hook in for input.
 */
static void *gre_input_save;
#endif

/*
 * driver functions
 */

static moduledata_t if_gre_mod = {
	"if_gre",
	if_gre_modevent,
	NULL
};

DECLARE_MODULE(if_gre, if_gre_mod, SI_SUB_PSEUDO, SI_ORDER_ANY);

static int
if_gre_modevent(module_t mod, int cmd, void *arg)
{
	int error = 0;

	switch (cmd) {
	case MOD_LOAD:
		if (loaded) {
			error = EEXIST;
		} else {
			gre_load();
			loaded = 1;
		}
		break;
	case MOD_UNLOAD:
		if (!loaded) { 
			printf("if_gre: gre_unload() called, but not loaded. "
			       "unload forced.\n");
		}
		error = gre_unload();
		break;
	default:
		error = EINVAL;
		break;
	}
	return error;
}

static int
gre_load(void)
{
	int error = 0;

	greunits->rm_type = RMAN_ARRAY;
	greunits->rm_descr = "configurable if_gre units";

	gre_install_input_hook();

	error = rman_init(greunits);
	if (error != 0)
		return error;

	error = rman_manage_region(greunits, 0, NGRE);
	if (error != 0) {
		printf("%s: greunits: rman_manage_region: Failed %d\n",
		       GREDEV, error);
		rman_fini(greunits);
		return (error);
	}
	if_clone_attach(&gre_cloner);

	loaded = 1;

	return error;
}

static int
gre_unload(void)
{
	int error = 0;

	gre_remove_input_hook();

	if_clone_detach(&gre_cloner);
	while (!TAILQ_EMPTY(&gres))
		gre_clone_destroy(&TAILQ_FIRST(&gres)->gre_if);
	error = rman_fini(greunits);
	if (error != 0)
		return error;

	return error;
}

/*
 * One major disadvantage of the BSD networking subsystem
 * is that individual protocols, and protocol families (domains),
 * cannot be dynamically loaded into the running kernel.
 *
 * For now, we can cheat and hook in on the RSVP
 * protosw instead. This is an extremely nasty hack,
 * and is not guaranrteed to work. You have been warned!
 *
 * RSVP is the chosen victim because it is in the base kernel
 * in 4.x, and it is hardly used except with VoIP or
 * MBONE applications.
 *
 * Alternatively, one can patch /usr/src/sys/netinet/in_proto.c,
 * and add the appropriate slots to the inetsw[] structure.
 *
 * The ideal solution is to use a dynamically growable contiguous
 * array, perhaps managed by a resource manager. This keeps
 * memory indirection low, and efficiency high.
 *
 * Everything implemented here is a hack to allow the GRE
 * driver to remain dynamically loadable/unloadable.
 */

static void
gre_install_input_hook(void)
{
	struct protosw *rawpsw = 0, *ppsw = 0;
	int s;

	s = splimp();
#ifdef GRE_EVIL_INPUT_HOOK_HACK
	printf("%s: WARNING: using IPPROTO_RSVP override method.\n", GREDEV);

	ppsw = pffindproto(AF_INET, IPPROTO_RSVP, SOCK_RAW);
	if (ppsw == 0)
		panic("%s: no handler for IPPROTO_RSVP\n", GREDEV);

	bcopy(ppsw, &rsvp_protosw_save, sizeof(struct protosw));
	bcopy(&gre_protosw, ppsw, sizeof(struct protosw));

	rawpsw = pffindproto(PF_INET, IPPROTO_RAW, SOCK_RAW);
	if (rawpsw == 0)
		panic("%s: no handler for IPPROTO_RAW\n", GREDEV);
	/* demote RSVP to being handled by the catchall */
	ip_protox[IPPROTO_RSVP] = rawpsw - inetsw;
#else
	ppsw = pffindproto(AF_INET, IPPROTO_GRE, SOCK_RAW);
	if (ppsw != 0) {
		gre_input_save = (void *) ppsw->pr_input;
		(void *) ppsw->pr_input = (void *) gre_input;
	} else {
		printf("%s: WARNING: protosw entry not found. "
		       "Driver will be write only.\n", GREDEV);
	}
#endif

	/* kick off GRE processing from ip_input() */
	ip_protox[IPPROTO_GRE] = ppsw - inetsw;
	splx(s);
}

static void
gre_remove_input_hook(void)
{
	struct protosw *rawpsw = 0, *ppsw = 0;
	int s;

	s = splimp();
	ppsw = pffindproto(AF_INET, IPPROTO_GRE, SOCK_RAW);
	rawpsw = pffindproto(PF_INET, IPPROTO_RAW, SOCK_RAW);
	if (rawpsw == 0)
		panic("%s: no handler for IPPROTO_RAW\n", GREDEV);

	/* demote GRE processing back to the raw IP handler */
	ip_protox[IPPROTO_GRE] = rawpsw - inetsw;

#ifdef GRE_EVIL_INPUT_HOOK_HACK
	/* promote RSVP processing back to its original handler */
	bcopy(&rsvp_protosw_save, ppsw, sizeof(struct protosw));
	ip_protox[IPPROTO_RSVP] = ppsw - inetsw;
#else
	(void *) ppsw->pr_input = (void *) gre_input_save;
	gre_input_save = 0;
#endif
	splx(s);
}

/*
 * create and initialize an individual gre clone instance.
 */
int
gre_clone_create(ifc, unit)
	struct if_clone *ifc;
	int *unit;
{
	struct resource *r;
	struct gre_softc *sc;

	if (*unit > NGRE)
		return (ENOSPC);

	if (*unit < 0) {
		r = rman_reserve_resource(greunits, 0, NGRE, 1,
			RF_ALLOCATED | RF_ACTIVE, NULL);
		if (r == NULL)
			return (ENOSPC);
		*unit = rman_get_start(r);
	} else {
		r = rman_reserve_resource(greunits, *unit, *unit, 1,
			RF_ALLOCATED | RF_ACTIVE, NULL);
		if (r == NULL)
			return (EEXIST);
	}

	sc = malloc (sizeof(struct gre_softc), M_GRE, M_WAITOK);
	bzero(sc, sizeof(struct gre_softc));

	/*
	 * initialize to reasonable defaults
	 */
	sc->gre_if.if_softc = sc;
	sc->gre_if.if_name = GRENAME;
	sc->gre_if.if_unit = *unit;
	sc->gre_if.if_mtu = GRE_DEFAULT_MTU;
	sc->gre_if.if_flags = IFF_POINTOPOINT | IFF_MULTICAST;
	sc->gre_if.if_ioctl = gre_ioctl;
	sc->gre_if.if_output = gre_output;
	sc->gre_if.if_type = IFT_OTHER;
	sc->gre_if.if_snd.ifq_maxlen = IFQ_MAXLEN;
	sc->gre_unit = r;
	sc->gre_proto = IPPROTO_GRE;
	if_attach(&sc->gre_if);
#if NBPFILTER > 0
#ifdef HAVE_OLD_BPF
	bpfattach(&sc->gre_if, DLT_NULL, sizeof(u_int));
#else
	bpfattach(&sc->gre_if.if_bpf, &sc->gre_if,
		  DLT_NULL, sizeof(u_int));
#endif
#endif
	TAILQ_INSERT_TAIL(&gres, sc, gre_link);
	return (0);
}

/*
 * destroy an individual gre clone instance.
 */
void
gre_clone_destroy(ifp)
	struct ifnet *ifp;
{
	struct gre_softc *sc = ifp->if_softc;
	int error;

	TAILQ_REMOVE(&gres, sc, gre_link);
	bpfdetach(ifp);
	if_detach(ifp);

	error = rman_release_resource(sc->gre_unit);
	if (error != 0)
		panic("if_gre: could not release resource\n");
	free(sc, M_GRE);
}

/*
 * handler for ioctl() calls to our interface driver.
 * this is called *after* the default handler, ifioctl(), which handles
 * most transactions. this is used by ifconfig(8).
 */
int
gre_ioctl(ifp, cmd, data)
	struct ifnet *ifp;
	u_long cmd;
	caddr_t data;
{
	struct gre_softc *sc = (struct gre_softc *) ifp;
	struct ifreq	*ifr = (struct ifreq *) data;
	struct in_ifaddr *ifa = (struct in_ifaddr *) data;
	struct in_aliasreq *iareq = (struct in_aliasreq *) data;
	struct sockaddr *sa = NULL;
	struct sockaddr_in si;
	int s = 0, error = 0;

	switch (cmd) {
	case SIOCSIFPHYADDR:
	case SIOCSLIFPHYADDR:
	case SIOCSIFPHYADDR_IN6:
		/* Check for IPv4 */
		if ((iareq->ifra_addr.sin_family != AF_INET) &&
		    (iareq->ifra_dstaddr.sin_family != AF_INET)) {
			printf("%s: not IPv4 endpoints\n", GREDEV);
			return EINVAL;
		}

		/* Check for null endpoint addresses */
		if ((iareq->ifra_addr.sin_addr.s_addr == INADDR_ANY) ||
		    (iareq->ifra_dstaddr.sin_addr.s_addr == INADDR_ANY)) {
			printf("%s: one or more null endpoints\n", GREDEV);
			return EINVAL;
		}

		sc->gre_src.s_addr = iareq->ifra_addr.sin_addr.s_addr;
		sc->gre_dst.s_addr = iareq->ifra_dstaddr.sin_addr.s_addr;
		break;
	/*
	 * query tunnel endpoints (used by ifconfig(8) amongst others).
	 */
	case SIOCGIFPSRCADDR:
	case SIOCGIFPSRCADDR_IN6:
		bzero(&si, sizeof(si));
		si.sin_len = sizeof(si);
		si.sin_family = AF_INET;
		si.sin_addr = sc->gre_src;
		bcopy(&si, &ifr->ifr_addr, sizeof(si));
		break;
	case SIOCGIFPDSTADDR:
	case SIOCGIFPDSTADDR_IN6:
		bzero(&si, sizeof(si));
		si.sin_len = sizeof(si);
		si.sin_family = AF_INET;
		si.sin_addr = sc->gre_dst;
		bcopy(&si, &ifr->ifr_addr, sizeof(si));
		break;
	/*
	 * configure endpoints for the tunnel.
	 * the encapsulation is ip-over-ip.
	 */
	case SIOCSIFADDR:
	case SIOCSIFDSTADDR:
		if (ifa->ia_ifa.ifa_addr->sa_family == AF_INET) {
			sa = ifa->ia_ifa.ifa_addr;
			if ((sc->gre_src.s_addr == INADDR_ANY) &&
			    (sc->gre_dst.s_addr == INADDR_ANY)) {
				sc->gre_src.s_addr =
					(IA_SIN(ifa))->sin_addr.s_addr;
				sc->gre_dst.s_addr =
					(IA_DSTSIN(ifa))->sin_addr.s_addr;
			}

			if ((sc->gre_src.s_addr != INADDR_ANY) &&
			    (sc->gre_dst.s_addr != INADDR_ANY)) {
				if (sc->gre_route.ro_rt != 0) {
					/* free cached route */
					RTFREE(sc->gre_route.ro_rt);
					sc->gre_route.ro_rt = 0;
				}
				/* mark as up */
				s = splimp();
				if_up(ifp);
				splx(s);
			}
		}
		break;
	case SIOCADDMULTI:
	case SIOCDELMULTI:
		error = EAFNOSUPPORT;
		break;
	/*
	 * XXX: Adapt MTU calculation to PMTU?
	 */
	case SIOCGIFMTU:
	case SIOCSIFMTU:
		{
			u_long	mtu;
			mtu = ifr->ifr_mtu;
			if (mtu < GRE_MTU_MIN || mtu > GRE_MTU_MAX) {
				return (EINVAL);
			}
			ifp->if_mtu = mtu;
		}
		break;
	case SIOCSIFFLAGS:
		/*
		 * both physical endpoints of the tunnel must be 
		 * configured for 'going up' to make sense.
		 */
		if ((sc->gre_dst.s_addr == INADDR_ANY) ||
		    (sc->gre_src.s_addr == INADDR_ANY)) {
			if (sc->gre_route.ro_rt != 0) {
				/* free cached route. */
				RTFREE(sc->gre_route.ro_rt);
				sc->gre_route.ro_rt = 0;
			}
			s = splimp();
			if_down(ifp);
			splx(s);
		}
		break;
	default:
		error = EINVAL;
	}

	return error;
}

/*
 * gif_input() dispatch GRE-tunneled datagram.
 * called from ip_protox[]/inetsw[] table.
 */
void
gre_input(m, off, proto)
	register struct mbuf *m;
	int off;
	int proto;
{
	register struct greip *gip = mtod(m, struct greip *);
	register struct ifqueue *ifq;
	register int iphlen = off;
	struct gre_softc *sc = 0;
	u_short flags;
	u_int af;
	int s;

	/*
	 * XXX: does this make sense if we now support ipfw?
	 */
	if (!gre_allow) {
		m_freem(m);
		return;
	}

	/*
	 * strip ip options, because we don't need them.
	 */
	if (iphlen > sizeof (struct ip)) {
		ip_stripoptions(m, (struct mbuf *)0);
		iphlen = sizeof(struct ip);
	}

	switch (proto) {
	case IPPROTO_GRE:
		/*
		 * XXX: length to m_pullup() by should really be determined
		 * by the options in effect on the ingress interface.
		 */
		if (m->m_len < sizeof(*gip)) {
			m = m_pullup(m, sizeof(*gip)+GRE_H_MAXLEN);
			if (!m)
				return;
		}
		if ((sc = gre_lookup(m, proto)) == NULL) {
			printf("%s: no matching tunnel found\n", GREDEV);
			m_freem(m);
			return;
		}
		m->m_pkthdr.rcvif = &sc->gre_if;

		/* determine packet options */
		flags = ntohs(gip->gi_flags);

		if ((gip->gi_flags & GRE_VER_MASK) != GRE_VERSION_0) {
			printf("%s: dropped non-RFC2784 packet\n", GREDEV);
			m_freem(m);
			return;
		}

		if (flags & GRE_RFC1701_MASK) {
			printf("%s: dropped RFC1701 packet with options\n",
			   GREDEV);
			m_freem(m);
			return;
		}

		off = iphlen + sizeof (struct gre_h);
		/* checksum present? */
		if (flags & GRE_CP) {
			u_int16_t ocksum, cksum;
			int pktlen = gip->gi_len;

			cksum = ((u_int16_t *)&gip->gi_ptype)[1];
			if (cksum) {
				/* point to GRE payload */
				m_adj(m, iphlen);
				ocksum = in_cksum(m, pktlen - iphlen);
				m_adj(m, -iphlen);
			}
			if (!cksum || ocksum) {
				sc->gre_if.if_ierrors++;
				m_freem(m);
				return;
			}
			/* point past cksum and reserved1 */
			off += sizeof(u_int16_t)*2;
		}

		/*
		 * currently, we only support IPv4.
		 */
		switch (ntohs(gip->gi_ptype)) {
		case ETHERTYPE_IP:
			/* pass to the IP queue */
			ifq = &ipintrq;
			af = AF_INET;
			break;
		default:
			sc->gre_if.if_ierrors++;
			m_freem(m);
			return;
		}
		/****/
		break;

	default:
		printf("%s: tunnel proto not IPPROTO_GRE, discarding\n",GREDEV);
		sc->gre_if.if_ierrors++;
		m_freem(m);
		return;
	}

	m_adj(m, off);

#if NBPFILTER > 0
	/*
	 * support bpf taps for tcpdump, including the faked AF header.
	 */
	if (sc->gre_if.if_bpf) {
		struct mbuf m0;

		m0.m_next = m;
		m0.m_len = 4;
		m0.m_data = (char *) &af;

		bpf_mtap(&sc->gre_if, &m0);
	}
#endif
	/* dispatch */
	s = splimp();
	if (IF_QFULL(ifq)) {
		IF_DROP(ifq);
		m_freem(m);
	} else {
		IF_ENQUEUE(ifq, m);
		/* 
		 * already in NETISR_IP at this point.
		 */
		/*schednetisr(NETISR_IP);*/
		sc->gre_if.if_ipackets++;
		sc->gre_if.if_ibytes += m->m_pkthdr.len;
	}
	splx(s);

	return;
}

/*
 * find tunnel matching source/destination/protocol tuple.
 */
struct gre_softc *
gre_lookup(m, proto)
	struct mbuf *m;
	u_int8_t proto;
{
	struct ip *ip = mtod(m, struct ip *);
	struct gre_softc *sc;

	TAILQ_FOREACH(sc, &gres, gre_link) {
		if ((sc->gre_dst.s_addr == ip->ip_src.s_addr) &&
		    (sc->gre_src.s_addr == ip->ip_dst.s_addr) &&
		    (sc->gre_proto == proto) &&
		    ((sc->gre_if.if_flags & IFF_UP) != 0))
			return (sc);
	}
	return (NULL);
}

int
gre_output(ifp, m, dst, rt)
	struct ifnet *ifp;
	struct mbuf *m;
	struct sockaddr *dst;
	struct rtentry *rt;
{
	int error = 0;
	struct gre_softc *sc = (struct gre_softc *) ifp;
	struct sockaddr_in *rdst = (struct sockaddr_in *)&sc->gre_route.ro_dst;
	struct greip gh;	/* XXX may change to alloca() in future*/
	struct ip *inp = NULL;	/* original IP header */
	u_short etype = 0;	/* ethernet type of inner payload */
	static int nestcnt = 0;	/* XXX wrap with mutex */

	/*
	 * Try to limit infinite recursion through misconfiguration.
	 */
	if (++nestcnt > gre_nest_limit) { 
		log(LOG_NOTICE,
		    "if_gre: gre_output() exceeded nestcnt limit (%d).\n",
		    nestcnt);
		m_freem(m);
		error = EIO;
		goto end;
	}

	/*
	 * pass packets back up to bpf for tcpdump.
	 */
#if NBPFILTER > 0
	if (ifp->if_bpf) {
		struct mbuf m0;
		u_int32_t af = dst->sa_family;

		m0.m_next = m;
		m0.m_len = 4;
		m0.m_data = (char *) &af;
#ifdef HAVE_OLD_BPF
		bpf_mtap(&sc->gre_if, &m0);
#else
		bpf_mtap(&sc->gre_if.if_bpf, &m0);
#endif
	}
#endif

	/*
	 * XXX: need to put IF_DROP() calls back in.
	 */

	if (sc->gre_proto == IPPROTO_GRE) {
		if (gre_allow == 0) {
			m_freem(m);
			error = EACCES;
			goto end;
		}

		switch(dst->sa_family) {
		case AF_INET:
			if (m && m->m_len < sizeof(struct ip)) {
				m = m_pullup(m, sizeof(struct ip));
				if (m == NULL) {
					error = ENOBUFS;
					goto end;
				}
			}
			inp = mtod(m, struct ip *);
			etype = ETHERTYPE_IP;
			break;

		default:
			m_freem(m);
			error = EAFNOSUPPORT;
			goto end;
		}

	} else {
		m_freem(m);
		error = EINVAL;
		goto end;
	}

	if (m == NULL) {
		error = ENOBUFS;
		goto end;
	}

	/*
	 * build encapsulating IP header
	 */
	bzero(&gh, sizeof(struct greip));
	gh.gi_i.ip_p   = sc->gre_proto;
	gh.gi_i.ip_src = sc->gre_src;
	gh.gi_i.ip_dst = sc->gre_dst;
	gh.gi_i.ip_tos = inp->ip_tos;
	gh.gi_i.ip_ttl = ip_defttl;
	gh.gi_i.ip_len = m->m_pkthdr.len + sizeof(struct greip);

	/*
	 * build GRE header
	 * XXX no support for additional options yet.
	 */
	if (sc->gre_proto == IPPROTO_GRE)
		gh.gi_ptype = htons(etype);

	/* prepend new IP + GRE header */
	M_PREPEND(m, sizeof(struct greip), M_DONTWAIT);
	if (m && m->m_len < sizeof(struct greip))
		m = m_pullup(m, sizeof(struct greip));
	if (m == NULL) {
		printf("if_gre: m_pullup() allocation failed\n");
		return ENOBUFS;
	}
	bcopy(&gh, mtod(m, struct greip *), sizeof(struct greip));

	/*
	 * check destination against desired route, and allocate
	 * new route if necessary.
	 */
	if (rdst->sin_addr.s_addr != sc->gre_dst.s_addr) {
		/* cache route doesn't match */
		rdst->sin_family = AF_INET;
		rdst->sin_len = sizeof(struct sockaddr_in);
		rdst->sin_addr = sc->gre_dst;
		if (sc->gre_route.ro_rt) {
			RTFREE(sc->gre_route.ro_rt);
			sc->gre_route.ro_rt = NULL;
		}
		/* set MTU back to default */
		ifp->if_mtu = GRE_DEFAULT_MTU;
	}

	if (sc->gre_route.ro_rt == NULL) {
		rtalloc(&sc->gre_route);
		if (sc->gre_route.ro_rt == NULL) {
			m_freem(m);
			return ENETUNREACH;
		}

		/* if it constitutes infinite encapsulation, punt. */
		if (sc->gre_route.ro_rt->rt_ifp == ifp) {
			m_freem(m);
			return ENETUNREACH;     /*XXX*/
		}
		/*
		 * calculate MTU over this interface
		 * XXX: does not use PMTU.
		 */
		ifp->if_mtu = sc->gre_route.ro_rt->rt_ifp->if_mtu
			- (sizeof(struct ip) + sizeof(struct gre_h));
	}

	/*
	 * send GRE packet.
	 */
	error = ip_output(m, NULL, &sc->gre_route, 0, NULL);
	ifp->if_opackets++;
	ifp->if_obytes += m->m_pkthdr.len;
end:
	nestcnt = 0;
	if (error)
		ifp->if_oerrors++;
	return (error);
}
